%let _metafolder = /Stored Processes/PW Metadata Explorer ;

/*
 Copyright 2012, 2018, 2019, Laurent de Walick, Frank Poppe.

See change history further down

This file is part of the PW Metadata Explorer.

    The PW Metadata Explorer is free software: you can redistribute it 
	and/or modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

    The PW Metadata Explorer is distributed in the hope that it will 
	be useful, but WITHOUT ANY WARRANTY; without even the implied 
	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the PW Metadata Explorer.  
	If not, see <https://www.gnu.org/licenses/>.

version 1.0.0 Frank Poppe
- Added %nrquote function around &_PROGRAM 

version 0.9.1 Frank Poppe 
- removed definition of macro var WorkPath (was not used anymore)
- renamed the macro variables for the Includes to Phase1, 2, 3.

version 0.9.2 Frank Poppe
- skipped

version 1.0.1 Frank Poppe
- added copyright notices

version 1.1.0 Frank Poppe
- Create macrovars holding _debug parm, to use in all calls
- facilitated easier debugging 
- merged all code together, stored in the metadata. It can now be exported into a SAS package (.spk)
	The application will notice when it has not been initialized.
	In that case the necesssary files will be exported from Notes in the package to files.
	Some .sas files will then be executed to produce the necessary support tables and files.

version 1.1.1 Frank Poppe
- On the Form-page, a search string without attribute-name will be used again ID, Name and Desc

version 1.2.0 Frank Poppe
- Changed the ObjectInfo window: added a breadcrumb-trail, make the object header a button
- Added a check on the AppVersion 
- A new session will no longer have the Root metadata type selected, added a warning if that is selected.

version 1.2.1 Frank Poppe
- Added icons to the objects
- Added a breadcrumb trail on the info page
- Made the PW_MDE_UI code such that is can run outside the SP to initialize all tables and files

*/

%let appversion = 1.2.1 ;
%let appVersionMin = 1.2.1; /* If utility files on disk are from an older version, they have to be refreshed */
%let dateCreated = '6oct2019'd ;

%let partLength = 250 ; /* for long attributes */


%global 
	PW_MDE_Type 
	PW_MDE_select 
	PW_MDE_object
	PW_MDE_jobInfo
	PW_MDE_attribute
	PW_MDE_update
	PW_MDE_code 
	PW_MDE_trail

	PW_MDE_prevType
	PW_MDE_prevSelect

	defaultFilter
	selected 

	PW_MDE_Phase1
	PW_MDE_Phase2
	PW_MDE_Phase3
	
	debugInput
	debugParm

;
%macro debugOnOff ;
/* Set parameters to retain  debug in <form> and <a href=> */
%if %symexist(_debug) %then %do ;
	%let debugInput=%str(<input type="hidden" name="_debug" value="&_debug" />);
	%let debugParm=_debug=&_debug ;
%end ;
%else %do ;
	%let debugInput=;
	%let debugParm=_debug=;
%end ;
%mend ;
%debugOnOff ;
%put &=debugInput ;
%put &=debugParm ;

%macro PW_MDE_fileContents ( fileref ) ;
/* if _debug then show filecontents for debugging */
%if %symExist(_debug) %then %do ;
data _null_ ;
infile &fileref ;
input ;
put _INFILE_ ;
run ;
%end ;
%mend ;

%let OMI_ALL=2**0 ;
%let OMI_FULL_OBJECT=2**1 ;
%let OMI_TEMPLATE=2**2 ;
%let OMI_ALL_SIMPLE=2**3 ;
%let OMI_INCLUDE_SUBTYPES=2**4 ;
 
%let OMI_XMLSELECT=2**7 ;
%let OMI_GET_METADATA=2**8 ;
%let OMI_MATCH_CASE=2**9 ;
 
%let OMI_SUCCINT=2**11 ;
 
%let OMI_NOFORMAT=2**26 ;


title ;
footnote ;

/* Retrieve the libref where the supporting tables are stored, and get attributes  */
/* There might be different versions around, so first determine the URI of the metadata folder */
/* Check if the _metafolder macrovariable exists (automatic in a SP) */

data _null_ ;
if not symexist ( '_metafolder' ) then do ;
	putlog 
		"NOTE: This code is the main part of the Stored Process 'PW Metadata Explorer'." / 
		"NOTE: Running it outside the Stored Process is possible, to set up files, after importing it in the metadata." /
		"ERROR%str(:) In that case the macrovariable _metafolder should be defined, and contain the metadata folder path to the Stored Process." ;
abort cancel ;
end ;
run ;

/*
First find out the URI of the metadatafolder.
We have to find the folder with the right name, in the right path 
(folder names can occur multiple times).

The obvious thing is to start at the begining, and then walk through the path
by getting each time alt the subfolders until the one with the right name is found.
But the metadata datastep functions used in a loop get very slow, 
and sometimes there are many subtrees.

Starting from the end is cumbersome, because a parent-child combination is not necessarily unique.
So we might have taken a few steps before discovering that we are walking the wrong path. 
Retracing the steps can be codes, but it gets a bit complicated.

The method applied here is start at the beginning, then getting all the folders that have 
the right name for the next subfolder, and then checking which of those has the right parent.
For the first folder the right parent is 'no parent'. 

*/

data _null_ ;
length
      uriCurrent uriParent lastParent folderName    
$ 200 
	metadataFolderID $ 17 
;
call missing (   uriCurrent , uriParent ) ; 
;

lastParent = '' ;
iFolder = 1 ;
folderName = scan ( "&_metafolder" , iFolder , '/' ) ;

do while ( folderName > '' ) ; /* no next folder in the path */

found = 0 ;
iChild = 1 ;
do while ( not found  ) ;
	nObj = metadata_getnobj ( cats ( "omsobj:Tree?@Name='" , folderName , "'" )  , iChild , uriCurrent ) ;
	/* find the parent of this folder */
	uriParent = '' ; /* reset because if no parent is found, it will still have the last one */
    rc = metadata_getnasn ( uriCurrent , "ParentTree" , 1 , uriParent ) ;

	if uriParent = lastParent then found = 1 ; 
	iChild + 1 ;
end;
lastParent = uriCurrent ;
iFolder  + 1 ;
folderName = scan ( "&_metafolder" , ifolder , '/' ) ;
putlog / iFolder= folderName= lastParent= / ;
end ;

metadataFolderId = scan ( uriCurrent  , 2 , '\' ) ;
call symputx ( 'metadatafolderID' , metadataFolderId , 'G' ) ;

run ;

%put &=metadatafolderId ;

/* Now find the SAS Library that has been mapped to our registered tables, and then its path 
We know now the metadata folder it is in */
data _null_ ;
length 
	uriLibrary uriDirectory LibraryPath libName
		$ 200
	libref 
		$ 8 
;
retain 
	name uriLibrary libref uriDirectory LibraryPath libName ''
;
/* Check on physical name, because the metadata name might have got a suffix
if a table with that name was already registered in another folder.
So first check if the table is member of the correct metadata folder 
*/
length uriTable uriTree $ 200 ;
call missing ( uriTable , uriTree ) ;

found = 0 ; iObj = 1 ; nTables  = 999 ;
do while (not found and iObj <= nTables ) ;
	nTables = metadata_getnobj ( "omsobj:PhysicalTable?@TableName='PW_MDE_ATTRIBUTES'"  ,iObj ,uriTable );
	rc = metadata_getnasn ( uriTable , "Trees" , 1 , uriTree ) ;
 	metadataFolderId = scan ( uriTree , 2 , '\' ) ;
	if metadataFolderId = "&metadataFolderId" then found = 1 ;
	iobj+1;
end ;

rc=metadata_getnasn ( uriTable , "TablePackage" , 1 , uriLibrary ) ;
put rc= uriLibrary= ;

rc=metadata_getattr(uriLibrary ,"Libref",libref);
put rc= libref= ;

rc=metadata_getattr(uriLibrary ,"Name", libname);
put rc= libname= ;

rc=metadata_getnasn (uriLibrary,"UsingPackages",1,uriDirectory);
put rc= uriDirectory= ;

rc=metadata_getattr(uriDirectory ,"DirectoryName",LibraryPath);
put rc= LibraryPath= ;

call symputX ( 'pw_mde_liburi' , uriLibrary , 'G' ) ;
call symputX ( 'pw_mde_libref' , libref , 'G') ;
call symputX ( 'pw_mde_libname' , libname , 'G') ;
call symputX ( 'pw_mde_libraryPath' , libraryPath , 'G') ;
run ;

%put &=pw_mde_liburi &=pw_mde_libName &=pw_mde_libref &=pw_mde_libraryPath;

/*libname pw_mde meta liburi="&pw_mde_liburi" metaout=data ;*/
/*libname pw_mde list ;*/
/*Meta engine does not work for ODS Path */

libname pw_mde "&pw_mde_libraryPath" ;
options fmtsearch = ( pw_mde ) ;

filename version "&pw_mde_libraryPath./PW_MDE_version.txt" ;

/* Check if the file PW_MDE_version exists */
%macro checkVersion ;
%global PW_MDE_versionOK ;
data _NULL_ ;
	versionFileExists = fileExist ( "&pw_mde_libraryPath./PW_MDE_version.txt" ) ;
	putLog versionFileExists= ;
	call symputx ( 'versionFileExists' , versionFileExists , 'G' ) ;
run ; 

%let PW_MDE_versionOK=0 ;
%if ( &versionFileExists ) %then %do ;
	data ;
	length sysscpl sysvlong4 $ 80 appVersion $ 20;
		infile version ;
		input sysscpl=$ sysvlong4=$ appversion=$ ;
		put sysscpl= sysvlong4= appversion= ;

		factor = 1e6 ;
		appVersionNum = 0 ;
		appversionMin = 0 ;
		do i = 1 to 3 ;
			num = input ( scan ( appVersion , i , '.' ) , 3. ) ;
			appVersionNum = appVersionNum + factor * num ;
			num = input ( scan ( "&appVersionMin" , i , '.' ) , 3. ) ;
			appVersionMin = appVersionMin + factor * num ;
			factor = factor / 1e3 ;
		end ;
		put appVersionNum= commax9. appVersionMin= commax9. ;
		if sysscpl = "&sysscpl"  and  sysvlong4 = "&sysvlong4" and appVersionNum >= appVersionMin then call symputx ( 'PW_MDE_versionOK' , 1 , 'G' )  ;
		else do ;
			call symputx ( 'pw_mde_OSversion' , sysscpl , 'G' ) ;
			call symputx ( 'pw_mde_SASversion', sysvlong4 , 'G' ) ;
			call symputx ( 'pw_mde_appversion', appVersion , 'G' ) ;
		end ;
	run ;
%end ;

%put &=PW_MDE_versionOK ;

%if &PW_MDE_versionOK ne 1 and &versionFileExists %then 
	%put &=pw_mde_OSversion &=pw_mde_SASversion &=pw_mde_appversion ;
%mend ;

%checkversion ;

ods path pw_mde.PW_MDE_templates (update ) sashelp.tmplmst ( read ) ;

ods escapechar = '~'  ;

options symbolgen mtrace ;
 
filename request temp encoding='utf-8';
filename result temp encoding='utf-8' ;
filename response temp encoding='utf-8' ;
filename xmlobj temp encoding='utf-8';
filename xslt temp encoding='utf-8';
filename xmlmap temp encoding='utf-8';
filename html temp encoding='utf-8';
filename content temp encoding='utf-8';

filename attribs "&pw_mde_libraryPath./PW_MDE_attributes.html"  encoding = 'utf-8';
filename assocs "&pw_mde_libraryPath./PW_MDE_associations.html" encoding = 'utf-8';


/* Check the pw_mde_select macro parm on quotes */ 
data _null_ ;
length string $ 1000 ;
if symexist ( 'pw_mde_select' ) then 
	string = symget ( 'pw_mde_select' ) ;
else
	string = '' ;
put string = ;
length = klength ( string ) ;
if length > 0 then do ;
	firstChar = Ksubstr ( string , 1 , 1 ) ;
	if firstChar ne '@' and firstChar ne '*'  then do ;

		/* generic attribute search, check for quotes */
		lastChar = ksubstr (  string, length , 1 ) ;
		putlog string= firstchar= lastchar= ;

		if not ( firstchar = "'" and lastChar = "'" ) 
		then do ;
			putlog 'add single quotes' ;
			string = quote ( trim ( string ) , "'" ) ;
		end ;
		put string= ;

		string = cat ( "@id ? " , ktrim ( string ) , " or @name ? " , ktrim(string) , " or @desc ? " , ktrim(string) ) ;
		put string= ;

	end ;
	else putlog 'string remains ' string ;
end ;
call symputx ( 'pw_mde_XML' , string , 'g' ) ;

run ;

%put pw_mde_XML: %nrbquote ( &pw_mde_XML ) ;

%macro PW_MDE_PrepUpdate ;
/* Shows information window, and creates the link that will execute the update */
%if ( &versionFileExists ) %then %do ;

proc stream outfile = content ; BEGIN<form
	action="/SASStoredProcess/do"
	method="POST"
>
<div class="panel panel-default">
<div class="panel-body" style="margin: 20px">
<p>
The supporting tables and tagset seem to have been produced on another system, or by a older version of this application. <br>
Your current system is &sysscpl., running SAS version &sysvlong4. 
and the version of this application is &appVersion., 
requiring files that have been produced by at least version &appVersionMin..<br>
The files have been produced on OS &PW_MDE_OSversion. using SAS version &pw_mde_SASversion. by application version &pw_mde_appVersion..
</p>
<p>
By clicking OK the tables and the tagset will be updated. This may take a while. 
<br>
On some servers this can run up to some minutes. Please be patient.
</p>
	<input type="hidden" name="_program" value="&_Program" />
	<input type="hidden" name="PW_MDE_update" value="yes" />
	&debuginput
<p>
	<button type="submit" class="btn btn-primary"><i>OK</i></button>
</p>
</div>
</div>
</form>
</div>

%end ;
%else %do ;

proc stream outfile = content ; BEGIN<form
	action="/SASStoredProcess/do"
	method="POST"
>
<div class="panel panel-default">
<div class="panel-body" style="margin: 20px">
<p>
The supporting tables and tagset do not yet exist, or are not complete.<br>
</p>
<p>
By clicking OK the tables and the tagset will be produced. This may take a while. 
<br>
On some servers this can run up to some minutes. Please be patient.
</p>
	<input type="hidden" name="_program" value="&_Program" />
	<input type="hidden" name="PW_MDE_update" value="yes" />
	&debuginput
<p>
	<button type="submit" class="btn btn-primary"><i>OK</i></button>
</p>
</div>
</div>
</form>
</div>

%end ;

;;;;
run ;

%PW_MDE_ResultPage ;

%mend /*PW_MDE_PrepUpdate*/;

%macro PW_MDE_ResultPage ;
/* Creates the general result window and display everything in the CONTENT fileref */

filename pwlogo "&pw_mde_libraryPath\PW_MDE_Logo.html" ;

proc stream outfile=_webout ; BEGIN<!DOCTYPE html>
<html lang="en">
<head>
  <title>PW Metadata Explorer</title>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css">
  <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js"></script>
  <script src="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/js/bootstrap.min.js"></script>
 
  <style> 
               /* Icon when the collapsible content is shown */ 
               .acc:before { 
               font-family: "Glyphicons Halflings"; 
               content: "\2212"; 
               float: left; 
               margin-right: 15px; 
               } 
               /* Icon when the collapsible content is hidden */ 
               .acc.collapsed:before { 
               font-family: "Glyphicons Halflings"; 
               content: "\002b"; 
               } 
				body {border-style: hidden;} ;
			   th {
				border: 1px dotted gray ;
				}

 
	code {
font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
background-color: #e6eeff;
color: #002633;
	}

	.assoc {  
	margin: 2px ;
	padding: 2px ;
/*	display: inline-block; */
    width: 110px;
	 text-align: center; 
	border: 1px solid gray; 
	border-radius: 0px; 

	pointer-events: none;
	font-size: 80%;  
	} 

	.object { 
	background-color: #B1D0E9l; 
	border: 2px solid #337ab7; 
	border-radius: 10px; 
	margin: 2px; 
 </style> 


</head>
<body>
 
<nav class="navbar navbar-inverse" style="margin-bottom: 0px;">
	<div class="container-fluid"  >
/*		<div class="navbar-text">*/
<a href="http://www.pwconsulting.nl" target="_blank">
<span class="navbar-text" >
&streamdelim readfile PWlogo ;
</span></a>

		<div class="navbar-header navbar-brand" >
PW Metadata Explorer
		</div>
		<p class="navbar-text">%nrstr(&nbsp;&nbsp;&nbsp;)</p>
		<ul class="nav navbar-nav" >
			<li class="active" ><a href="//&_srvname:&_srvport/&_url?_PROGRAM=&_program%nrstr(&)&debugparm%nrstr(&pw_mde_prevtype=)&pw_mde_type%nrstr(&pw_mde_prevselect=)&pw_mde_select"><span class="glyphicon glyphicon-search"></span> New Search</a></li>
		</ul>
		<p class="navbar-text navbar-right">Version &appversion%nrstr(&nbsp;)</p>
		
	</div>
</nav>

</nav>

<div class="container-fluid">  
&streamdelim readfile content  ;
</div>

<hr>
 
</body>
</html>
;;;;

%mend /* PW_MDE_ResultPage */ ;

%macro PW_MDE_ExecUpdate ;
/* Initialze the application */

filename textStor temp encoding='utf-8';
filename htmlMap  temp encoding='utf-8';
filename sasCode  temp encoding='utf-8';

%put &=metadataFolderID ;

filename request temp ;

%let flag=%eval( 0 
/*+ &OMI_ALL*/
/*+ &OMI_FULL_OBJECT*/
+ &OMI_TEMPLATE
/*+ &OMI_ALL_SIMPLE */
/*+ &OMI_INCLUDE_SUBTYPES */
+ &OMI_XMLSELECT 
+ &OMI_GET_METADATA 
/* + &OMI_MATCH_CASE */
 
+ &OMI_SUCCINT 
/*+ &OMI_NOFORMAT */
);
 
/* Create a request to get all Documents within the Folder */
proc stream outfile=request ; BEGIN
 
<GetMetadataObjects>
<ReposId>$METAREPOSITORY</ReposId>
<Type>Document</Type>
<Objects/>
<Ns>SAS</Ns>
<Flags>&flag</Flags>
<Options>
 
<Templates>
 
<Document Name="" Desc="" >
<Notes/>
</Document>
 
<TextStore  StoredText="" TextRole="" TextType="" />
 
</Templates>
 
<XMLSelect search="Document[Trees/Tree[@Id='&MetadatafolderID']]" />
 
</Options>
</GetMetadataObjects>

;;;;

filename txtstors temp  ;
proc metadata in=request out=txtstors ;
run ;

/* Define an XML map to get a numbered list of document names from the XML with all the textstores*/
filename docmap temp ;

proc stream outfile = docmap ; BEGIN<?xml version="1.0" encoding="windows-1252"?>
<!-- ############################################################ -->
<!-- 2019-09-04T14:15:21 -->
<!-- SAS XML Libname Engine Map -->
<!-- Generated by XML Mapper, 903000.0.0.20110518190000_v930 -->
<!-- ############################################################ -->
<SXLEMAP name="AUTO_GEN" version="2.1">

    <NAMESPACES count="0"/>

    <!-- ############################################################ -->
    <TABLE description="Document" name="Document">
        <TABLE-PATH syntax="XPath">/GetMetadataObjects/Objects/Document</TABLE-PATH>

        <COLUMN class="ORDINAL" name="NumDocument">
            <INCREMENT-PATH beginend="BEGIN" syntax="XPath">/GetMetadataObjects/Objects/Document</INCREMENT-PATH>
            <TYPE>numeric</TYPE>
            <DATATYPE>integer</DATATYPE>
        </COLUMN>

        <COLUMN name="Id">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Document/@Id</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>17</LENGTH>
        </COLUMN>

        <COLUMN name="Name">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Document/@Name</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>38</LENGTH>
        </COLUMN>

    </TABLE>

</SXLEMAP>
;;;;

libname txtstors xmlv2 xmlmap = docmap ;


data document ;
set txtstors.document ;
put _ALL_ ;
run ;


filename xslt temp ;
/* Define XSLT to move the StoredText attribute values (can be over 32767 bytes) to a file*/
proc stream outfile=xslt; BEGIN<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="text" />

 <xsl:template match="/">
  <xsl:apply-templates select="/GetMetadataObjects/Objects/Document/Notes/TextStore" />
 </xsl:template>


 <xsl:template match="TextStore">
  <xsl:value-of select="@StoredText" />
 </xsl:template>

</xsl:stylesheet>
;;;;

filename stortxt temp ;
/* Apply the XSLT */
Proc XSL in=txtstors out=stortxt xsl=xslt ;
run ;

/* Need XML element around everyting to make it valid */
filename docsXML temp ;

Proc Stream outfile=docsXML ; BEGIN
<docs>
&streamdelim readfile stortxt ;
</docs>
;;;;

/* Create an XML-map to extract all the lines from the surrounding <p>...</p> tags 
Number per document to match with the document names extracted previously  */
filename docLines temp ;
proc stream outfile=doclines ; BEGIN
<?xml version="1.0" encoding="windows-1252"?>
<!-- ############################################################ -->
<!-- 2019-09-04T14:43:38 -->
<!-- SAS XML Libname Engine Map -->
<!-- Generated by XML Mapper, 903000.0.0.20110518190000_v930 -->
<!-- ############################################################ -->
<SXLEMAP name="AUTO_GEN" version="2.1">

    <NAMESPACES count="0"/>


    <!-- ############################################################ -->
    <TABLE description="Separate lines for all textstores" name="TextStoreLines">
        <TABLE-PATH syntax="XPath">/docs/html/body/p</TABLE-PATH>

        <COLUMN class="ORDINAL" name="numDocument">
            <INCREMENT-PATH beginend="BEGIN" syntax="XPath">/docs/html</INCREMENT-PATH>
            <TYPE>numeric</TYPE>
            <DATATYPE>integer</DATATYPE>
        </COLUMN>

        <COLUMN name="line">
            <PATH syntax="XPath">/docs/html/body/p</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>32767</LENGTH>
        </COLUMN>

    </TABLE>

</SXLEMAP>
;;;;

libname docsXML xmlv2 xmlmap = doclines ;


/* Now combine the document names with the lines */
proc sql ;
create table docs as 
select d.numDocument , line , Id , cats ( "&pw_mde_libraryPath/" , Name ) as PathName 
from txtStors.Document as d join docsXML.TextStoreLines as l
on d.numDocument = l.numDocument 
;

/* And write the lines to the separate documents */
data _null_ ;
length includeList $ 999 ;
retain includeList '' ;

set docs ;
by numDocument ;

if first.numDocument then do ;

	filename = kScan ( pathname , -2 , '\/.'  ) ;
	putlog pathname=  filename= ;

	if fileName =: 'PW_MDE_INI_' then do ;

		includeList = cat ( trim(includeList) , ' %include ''' , trim(pathname) , ''';' ) ;
		putlog includeList= ;
		call symputx ( 'includeList' , includeList , 'G' ) ;

	end ;

end ;

file out filevar = pathName ;
put  line ;
run ;


options symbolgen source2 mprint ;

&includeList
;

data _null_ ;
	file version ;
	put "sysscpl=&sysscpl. sysvlong4=&sysvlong4. appVersion=&appVersion." ;
run ;

%mend /* PW_MDE_ExecUpdate */ ;

%macro defaultForm ;
/* Utility to setup defaults for the form */

%if "&pw_mde_prevselect" > ""  %then %let defaultFilter = value="&pw_mde_prevselect" ;
%else %let defaultFilter = placeholder="Filter, e.g. @name?'some text' (attribute name contains 'some text')" ;

%if "&pw_mde_prevType" > "" %then %let selected =&pw_mde_prevtype ;
%else %let selected=;

%mend /* defaultForm */;

%macro HTML_select ( selectName , attributes , table , valueVar , DescrVar , valueSel , size, header) ;
/* Utility to create a select (i.e. dropdown) from a SAS table 
Parameters:
	selectName : the name the select object will have
	attributes : any other attributes to put inside the <select ... > tag
	table      : the SAS table that contains the list of options
	valueVar   : the variable in the table that contains the value that has to be returned when option is selected
	descrVar   : the variable that contains the value that has to be displayed
	valueSel   : the option value that initially should be marked as selected 
	size       : the value for the size attribute (defaults to 1)
	header     : start the list with a non-active header
*/
%if size= %then %let size=1 ;
%put &=header;
<select id="&selectName" name="&selectName" size="&size" &attributes >
%if ( "&header" > "" ) %then %do ;
	<option value="" disabled selected>&header</option>
%end ;
%let dsid = %sysfunc ( open ( &table ) ) ;
%let valueNum = %sysfunc ( varnum ( &dsid , &valueVar )) ;
%let descrNum = %sysfunc ( varnum ( &dsid , &DescrVar )) ;
%let rc = 0 ;
%do %until ( &rc ne 0 ) ;
	%let rc = %sysfunc ( fetch ( &dsid ) ) ;
	%if ( &rc = 0 ) %then %do ;
		%let value = %sysfunc ( getvarc ( &dsid , &valueNum )) ;
		%let descr = %sysfunc ( getvarc ( &dsid , &descrNum )) ;
		%let selected=;
		%if &value = &valueSel %then %let selected = selected ;
		<option value="&value" &selected>&descr</option>
	%end ;
%end ;
%let rc = %sysfunc ( close ( &dsid ) ) ;
</select>
%mend ;


%macro UpdateTableTagset ;
%if "&PW_MDE_update" = "yes" %then %PW_MDE_ExecUpdate ;
%mend ;

%macro PW_MDE_Form ;

/* Will be 'yes' if we come from the Information window that explains that updates are necessary */
%if "&PW_MDE_update" = "yes" %then %PW_MDE_ExecUpdate ;

/*%UpdateTableTagset ;*/

%let PW_MDE_trail =;

filename types temp encoding='utf-8';
filename typemap temp encoding='utf-8';

%defaultForm ;
%put &=defaultFilter &=selected ;

options symbolgen ;

title ;
footnote ;

options nosymbolgen nomprint nomtrace ;

proc stream outfile=content ; BEGIN<script>
var currentType = "&selected" ;
function showObjectHelp(selectedType) {
/*alert ( 'selectedType:' + selectedType ) ;*/
if ( selectedType == 'Root' ) { 
	alert ( 'Selecting the root object type will select ALL objects. Be sure to add an adequate filter' ) ;

/*Dit krijg ik nog niet werkend*/
/*	$("#pw_mde_select").text("Filter") ;*/
}

	var x = document.getElementsByClassName ( 'objectType_' + currentType ) ;
	var i;
	for (i = 0; i < x.length; i++) {
 	   x[i].style.display = "none" ; 
	}
		
	x = document.getElementsByClassName ( 'objectType_' + selectedType ) ;
	for (i = 0; i < x.length; i++) {
 	   x[i].style.display = "table-cell" ; 
	}
	currentType = selectedType ;
}
</script>

<script>
$(document).ready(function(){
showObjectHelp(currentType);
});
</script>

<div class="row">
    <div class="col-lg-2" ></div>
    <div class="col-lg-6" >


<form
	action="/SASStoredProcess/do"
	method="POST"
>
<div class="panel panel-default">
<div class="panel-body" style="margin: 20px">
<p>
Select an object type, and optionally a filter.<br>
This will return a list of objects.
%html_select ( pw_mde_Type , %str(class="form-control" onchange="javascript:showObjectHelp(value)" ) , 
pw_mde.pw_mde_types ,Type_id ,  type_desc ,  &selected , 1 , Please select an object type)
Items marked with an asterisk * have subtypes.<br>
Selecting the 'Metadata Root Abstract Type' selects <i>all</i> types.
<br>
</p>
<p>
<input type="text"  class="form-control"  name="pw_mde_select" id="pw_mde_select"
&defaultFilter 
/*value="&defaultFilter" */
/*placeholder="&defaultFilter"*/
/>
If left empty all objects of the requested type will be shown.<br>
Search on specific attributes using syntax like <code>@name?'some text'</code> <i>(<code>?</code> is the </i>contains<i> operator).</i><br>
When only a string is given, the search will be for objects having that string in <code>id</code>, <code>name</code> or <code>descr</code> attributes.</br>
The tables below give the valid attributes and associations for the selected object type.
</p>
	<input type="hidden" name="_program" value="&_Program" />
	&debugInput
<p>
	<button type="submit" class="btn btn-primary"><i>Go</i></button>
</p>
</div>
</div>
</form>

</div>

<div class="col-lg-4" ></div>
</div>

<div class='row'>

    <div class="col-lg-1" ></div>
    <div class="col-lg-5" >
		<h4>Valid attributes</h4>
&streamdelim readfile attribs ;
	</div>
    <div class="col-lg-5" >
		<h4>Valid associations</h4>
&streamdelim readfile assocs ;


	</div>
    <div class="col-lg-1" ></div>

</div>


<div class="row">
    <div class="col-lg-2" ></div>
    <div class="col-lg-6" >

<div class="panel panel-default">


<div class="panel-body" style="margin: 20px">
  <ul class="nav nav-pills ">
    <li class="active"><a data-toggle="pill" href="#tab1">Filter help</a></li>
    <li><a data-toggle="pill" href="#tab2">Tips</a></li>
    <li><a data-toggle="pill" href="#tab3">About</a></li>
  </ul>
  
  <div class="tab-content">
    <div id="tab1" class="tab-pane fade in active">
      <h3>How to construct a filter</h3>
      <p>
		A filter can be criteria on <i>attributes</i> or on <i>associations</i>, or on combinations of those. 
		Criteria on attributes are specified in the form %nrstr(<code>@<i>&lt;attribute-name&gt;&lt;operator&gt;value</i></code>), e.g. <code>@name='some text'</code>. 
		Operators can be <code>=</code>, <code>=:</code>, <code>?</code> (contains), <code>NE</code> (not equal), 
		<code>GE</code>, <code>GT</code>, <code>LE</code> and <code>LT</code>. 
		You can put <code>NOT</code> in front of criteria, with parentheses around the criteria.
		<br>
		Criteria on associations are given after an attribute criteria. A series of associations (a path) can be expressed by separating associations with a slash <code>/</code>.
		A path always starts with an association name, and then alternates with the object type.
		<br>
		In the application a filter that does <i>not</i> start with the attribute indicator <code>@</code>
		is translated into a search on the given text in the <code>@Name</code>, <code>@Desc</code> and <code>@Id</code> attributes.
	</p>
    </div>
    <div id="tab2" class="tab-pane fade">
      <h3>Tips and tricks, shortcuts</h3>
      <p>To find Generated Transformations search for <b>Prototypes</b> with <code>@PublicType='GeneratedTransform'</code>. 
		This will include all Generated Transformation delivered by SAS. 
		They will all have a <code>MetadataCreated</code> attribute reflecting the date and time of installation of the SAS system.
		<i>User</i>Generated Transformations can be selected by adding a datetime filter: 
		<code>@PublicType='GeneratedTransform' and @MetadataCreated GT '01nov2017:00:00:00'</code>. 
		Note that the day should be given with two digits, and that hours, minutes and seconds all have to be specified, also with two digits.
		Optionally the seconds may be given with one decimal. Alternatively a datetime value can be given as a number, between quotes.
	</p>
    </div>
    <div id="tab3" class="tab-pane fade">
      <h3>About this tool</h3>
      <p>Created by <i>Laurent de Walick</i> and <i>Frank Poppe</i>, from <a href="https://www.pwconsulting.nl"><b>PW Consulting</b></a></p>
	  <p>The application is released under the terms of Version 3 of the GNU General Public License as published by
    the <a href="https://www.gnu.org/licenses/licenses.html">Free Software Foundation</a>. 
    </div>
  </div>
</div>

</div>
</div>

<div class="col-lg-4" ></div>

</div>

;;;;

%PW_MDE_ResultPage ;

%mend ;


%macro PW_MDE_objects ;

 
%let gFlag=%eval( 0 
/*+ &OMI_ALL*/
/*+ &OMI_FULL_OBJECT*/
+ &OMI_TEMPLATE
/*+ &OMI_ALL_SIMPLE */
+ &OMI_INCLUDE_SUBTYPES 
+ &OMI_XMLSELECT 
+ &OMI_GET_METADATA 
/* + &OMI_MATCH_CASE */
 
+ &OMI_SUCCINT 
+ &OMI_NOFORMAT 
);
 

proc stream outfile=request ; BEGIN<GetMetadataObjects>
<Reposid>$METAREPOSITORY</Reposid>
<Type>&pw_mde_type</Type>
<Objects/>
<NS>SAS</NS>
<Flags>&gFlag</Flags>
<Options>
<Templates>
	<Root Id="" Name="" Desc="" MetadataCreated="" MetadataUpdated=""/>
</Templates>
<XMLSELECT search="&pw_mde_XML"/>
</Options>
</GetMetadataObjects>
;;;;


%PW_MDE_FileContents ( request ) ;
 		   				 
proc metadata in=request out=result ;
run;

/* Store return code and error message for future reference */
%let MDsyserr = &syserr ;
%let MDsysErrorText = &syserrortext ;
 
/*
For all objects (i.e. every tag between <Objects> and </Objects>),
transform it to a tag <ObjectType> with the original tag as Type attribute.
This makes it possible to process the XML with a single XML-map
*/

proc stream outfile = xslt ; BEGIN<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes"/>
<xsl:strip-space elements="*"/>
<!-- identity transform -->
<xsl:template match="@*|node()">
    <xsl:copy>
        <xsl:apply-templates select="@*|node()"/>
    </xsl:copy>
</xsl:template>
<xsl:template match="Objects/*">
    <Object>
        <xsl:attribute name="ObjectType">
            <xsl:value-of select ="name(.)"/>
        </xsl:attribute>
        <xsl:apply-templates select="@*|node()"/>
    </Object>
</xsl:template>
</xsl:stylesheet>
;;;;

proc xsl
   in=result 
   xsl=xslt
   out=xmlobj
;
run;
 
proc stream outfile=XMLmap ; BEGIN<?xml version="1.0" encoding="windows-1252"?>
<!-- ############################################################ -->
<!-- 2018-01-05T20:39:06 -->
<!-- SAS XML Libname Engine Map -->
<!-- Generated by XML Mapper, 904300.0.0.20150204190000_v940m3 -->
<!-- ############################################################ -->
<SXLEMAP name="ReadObjects" version="2.1">
 
    <NAMESPACES count="0"/>
 
    <!-- ############################################################ -->
    <TABLE name="Objects">
        <TABLE-PATH syntax="XPath">/GetMetadataObjects/Objects/Object</TABLE-PATH>
 
        <COLUMN name="Object_ObjectType">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@ObjectType</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>40</LENGTH>
        </COLUMN>
 
        <COLUMN name="Object_Id">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@Id</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>17</LENGTH>
        </COLUMN>
 
        <COLUMN name="Object_Name">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@Name</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>60</LENGTH>
        </COLUMN>
 
        <COLUMN name="Object_Desc">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@Desc</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>200</LENGTH>
        </COLUMN>
 
        <COLUMN name="Object_MetadataCreated">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@MetadataCreated</PATH>
            <TYPE>numeric</TYPE>
            <DATATYPE>datetime</DATATYPE>
            <FORMAT ndec="0" width="30">NLDATM</FORMAT>
            <INFORMAT ndec="2" width="18">F</INFORMAT>
        </COLUMN>
 
        <COLUMN name="Object_MetadataUpdated">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/Object/@MetadataUpdated</PATH>
            <TYPE>numeric</TYPE>
            <DATATYPE>datetime</DATATYPE>
            <FORMAT width="30">NLDATM</FORMAT>
            <INFORMAT ndec="2" width="12">F</INFORMAT>
        </COLUMN>
 
    </TABLE>
 
</SXLEMAP>
;;;;


libname xmlobj xmlv2  xmlmap=xmlmap ;

proc sort  data = xmlobj.objects out = objects ;
by object_objectType ;
run ;

%let dsid = %sysfunc( open ( objects ) ) ;
%let nObjects = %sysfunc ( attrn ( &dsid , NOBS ) ) ;
%put &=nObjects ;
%if ( &MDsysErr > 0 ) %then	%PW_MDE_Error ;
%else %if ( &nObjects = 0 ) %then %PW_MDE_NoObs ;
%else %PW_MDE_ShowObjects ;
%let syscc = 0 ;
%let rc = close ( &dsid ) ;

%mend ;

%macro PW_MDE_ShowObjects ;

ods _ALL_ close ;
filename objects temp encoding='utf-8' ;

ods tagsets.bootstrap file = objects  ;

proc report nowd data = objects ;
by Object_ObjectType ;
columns object_ObjectType object_id object_name object_desc ;
define object_ObjectType / noprint ; /* must be available to define the link */
define object_id / "Id" style(column)=[cellwidth = 20em textalign=r] ;

define object_name / "Name" ;
define object_Desc / "Desc" ;
compute object_id ; 
	call define ( _COL_ , "URL" , 
		cats ( "//&_srvname:&_srvport/&_url?_program=&_program" , '&' , "&debugParm" , '&pw_mde_type=' , object_objectType , '&pw_mde_object=' , object_id  ) 
	) ;
endComp ;

run ;

ods tagsets.bootstrap close ;


proc stream outfile=content ; BEGIN<div class="row" style="padding: 6px;">
<p>Results for the query on object type <b><tt>&pw_mde_type.</tt></b>%nrstr(&ensp);
%if "&pw_mde_select" > "" %then %let PW_MDE_filterclause = with filter <b><tt>%superQ(PW_MDE_select)</tt></b> ;
%else %let PW_MDE_filterClause = without a filter;
%superq(PW_MDE_filterClause)
</p><p>
This returned in total <b>&nObjects. objects</b>.
</p>
</div>

&streamdelim readfile objects ;
;;;;

%PW_MDE_ResultPage ;

%mend ;

%macro PW_MDE_Error ;

%if "&PW_MDE_select" > "" %then %let PW_MDE_filterclause = with filter <b><code>&PW_MDE_select</code></b> ;
%else %let PW_MDE_filterClause = without a filter;

proc stream outfile=content ; BEGIN<h2>Error in de query</h2>
<p>
Your query for objects of type <b><code>&pw_mde_type</code></b> &PW_MDE_filterClause. seems to generate an error.
</p><p>
The system returned an return code of &MDsyserr.
</p><p>
The error message was <b><code>&MDsysErrorText;</code></b>.
</p><p>
Please try again.
;;;;

%PW_MDE_ResultPage ;

%mend ;

%macro PW_MDE_NoObs ;

%if "&PW_MDE_select" > "" %then %do ;
	%let PW_MDE_filterclause = with filter <b><code>&PW_MDE_select</code></b> ;
	%let PW_MDE_tip = Maybe the filter is too strict.;
%end ;
%else %do ;
	%let PW_MDE_filterClause = without a filter;
	%let PW_MDE_tip = There seem to be no objects of that type, or you are not authorized to see them.;
%end ;
proc stream outfile=content ; BEGIN<h2>No results</h2>
Your query for objects of type <b><code>&pw_mde_type</code></b> &PW_MDE_filterClause. 
did not generate an error but returned no observations.
</p><p>
&PW_MDE_tip
</p><p>
Please try again.
;;;;

%PW_MDE_ResultPage ;

%mend ;

%macro PW_MDE_Info ;

%let flag = 1;

/* PWMET-15 
If the Encoding is left out, the data step logic that reads associations and attributes fails
The UTF-8 at the beginning are the problem.
Preferably this data step is to replaced by Proc XSL
*/
filename reqObj temp  encoding='utf-8';
filename resObj temp  encoding='utf-8' ;
filename resObjA temp  encoding='utf-8' ;

data _null_;
file reqObj lrecl=32767;
 put '<GetMetadata>';
 put '<Metadata>';
 put "<&pw_mde_type Id=" '"' "&pw_mde_object" '" Name="" Desc="">';
 put "</&pw_mde_type>";
 put '</Metadata>';
 put '<NS>SAS</NS>';
 put "<Flags>&flag.</Flags>";
 put '<Options/>';
 put '</GetMetadata>';
;;;;
run;

proc metadata
 in=reqObj
 out=resObj
;
run;

%PW_MDE_fileContents ( resObj ) ;

/* If a link to an attribute has to inserted, because the value is too long, 
select the first and last parts, with a marker in between.
The link that has to be inserted in stead of the marker can be created later */
proc stream outfile = xslt ; BEGIN<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" >
<xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes"/>
<xsl:template match="@*|node()">
  <xsl:copy>
   <xsl:apply-templates select="@*|node()"/>
  </xsl:copy>
</xsl:template>
<xsl:template match="@*" >
  <xsl:attribute name="{name()}" >
   <xsl:choose>
    <xsl:when test="string-length(.) > 1000">
      <xsl:value-of select="substring(.,1,&partLength)" />
      <xsl:value-of select="' [{PW_MDE_AttributeLink}] '"/>     
      <xsl:value-of select="substring(.,string-length(.)-&partLength,&partLength)" />
	</xsl:when>
    <xsl:otherwise>
     <xsl:value-of select="." />
    </xsl:otherwise>
   </xsl:choose>
  </xsl:attribute>
</xsl:template>
</xsl:stylesheet>
;;;;

%let rc = %sysfunc ( fdelete ( resobja ) ) ;
%put &=rc ;
%let loop = 0 ;
%put STARTING LOOP XSL ;
%do %while (%sysfunc(fexist(resobja)) = 0 and &loop. < 10 );
%let loop = %eval(&loop. + 1);
%put &=loop;
Proc XSL in=resobj out=resobjA xsl=xslt ;
run ;
%end;

%PW_MDE_fileContents ( resObjA ) ;


data readXml (keep=level tag compress=yes);
 attrib
  level length=8
  tagIn length=$32767;
 retain level 0 ;
 infile resObjA recfm=n lrecl=32767 dlm="<" firstobs =  2 ;
 input tagIn;
 putlog tagIn= ;
 tag = '<' || tagIn;
 if tag =: '<?xml' then do ; putlog 'drop' ; return ; end ;
 if index(tag,'</') gt 0 then level = level - 1;
 output;
 if index(tag,'/>') gt 0 then level = level;
 else if index(tag,'</') eq 0 then level + 1;
 put level= ;
run;

data attributes (keep=objectType objectId objectName objectDesc attributeName attributeValue attributeOrder compress=yes)
     associations (keep=objectId associationClass associationType associationId associationName associationDesc compress=yes)
	objinfo ( keep = objectType objectPublicType objectName  ObjectDesc   objectId   ) 
;
 set readXml end=eof;
 length
  objectType $64
  objectId $17
  objectType objectName $64
  objectDesc $256
  associationClass $64
  associationId $17
  associationType $64 associationName $200
  associationDesc $256
  attributeName $200
  attributeValue $32767
	objectPublicType $ 40   ;

 ;
 retain associationClass objectType objectId
	objectPublicType objectName  ObjectDesc    ;

;
 tag = tranwrd(tag,'""','" "');
 if (level = 3 and index(tag, '</') eq 0) then do;
  objectType = substr(scan(tag,1,' '),2);
  tag = substr(tag,length(objectType)+3);
  outp = 0;
  do while (index(tag,'=') ne 0);
   attributeName = scan(tag,1,'=');
   attributeValue = scan(tag,2,'"');
   tag = substr(tag,length(attributeName) + length(attributeValue) + 5); 
   attributeName = kstrip ( attributeName ) ;
   if lowcase(attributeName) = 'id' then do;
    objectId = attributeValue; attributeOrder = 4 ;
   end;
   else if lowcase(attributeName) = 'name' then do ; objectName = attributeValue ; attributeOrder = 2 ; end ;
   else if lowcase(attributeName) = 'desc' then do ; objectDesc = attributeValue ; attributeOrder = 3 ; end ;
   else if lowcase(attributeName) = 'publictype' then do ; objectPublicType = attributeValue ; attributeOrder = 1 ; end ;
   else do;
	attributeOrder = 10 ;
    outp = 1;
   end;
   	attributeName = left ( attributeName ) ;
    output attributes;
  end;
  if (outp = 0) then output attributes;
 end;
 else if level = 4 then
  associationClass = substr(scan(tag,1,'>'),2);
 else if level = 5 then do;
  associationId = '';
  associationName = '';
  associationDesc = '';
  associationType = scan(tag,1);
  tag = substr(tag,length(associationType)+3);
  outp = 0;
  do while (index(tag,'=') ne 0);
   attributeName = scan(tag,1,'=');
   attributeValue = scan(tag,2,'"');
   if lowcase(attributeName) = 'id' then associationId = attributeValue;
   else if lowcase(attributeName) = 'name' then do ;
		associationName = attributeValue;
		/* check if already in trail, and if so add info */
		if index (  "&pw_mde_trail" , associationID ) > 0 then 
		associationName = cat ( '<div>' , trim(associationName) , ' <i><b>(already in this trail)</b></i></div>' )
	;

	end ;
   else if lowcase(attributeName) = 'desc' then associationDesc = attributeValue; 
   tag = substr(tag,length(attributeName) + length(attributeValue) + 5);
  end;
  output associations;
 end;

if eof then output objinfo ;
run;

proc sort data=associations EQUALS ;
 by associationClass associationType ; 
run;

proc sort data = attributes EQUALS ;
by attributeOrder ;
run ;

data attributes  ;
set 
	attributes 
;
length 
	partStart partEnd $ &partLength  
;

attributeMark = '[{PW_MDE_AttributeLink}]' ;
attrMarkLength = klength ( attributeMark ) ;
attributelink = kindex ( attributeValue , attributeMark ) ;
putlog attributeName= attributelink= ;

attributeValue = cats ( '<div>' , attributeValue , '</div>' ) ;

if attributeLink > 0 then do ;
/* There is marker in stead of which a link to a new page/tab has to be created */
	partStart = urldecode ( ksubstr ( attributeValue , 1, attributeLink - 1 ) ) ;
	partEnd   = urldecode ( ksubstr ( attributeValue , ( attributeLink + attrMarkLength + 6 ) ) ) ;
	putlog partStart= / partEnd= ;
	attributeValue = cats ( 
		partStart ,
		'&nbsp;<a class="btn btn-primary" role="button" target="_blank" href="' ,
		"//&_srvname:&_srvport/&_url?_PROGRAM=&_program%nrstr(&)&debugparm%nrstr(&pw_mde_type=)&pw_mde_type%nrstr(&pw_mde_object=)&pw_mde_object%nrstr(&pw_mde_attribute=)" ,
		attributeName , 
		'">Content</a>&nbsp;' ,
		partEnd ,
		'<div style="text-align: left;" class="alert alert-info"><i>Value is too long to display here, only begin and end are shown. Click for full content.' , 
		'<br></i>(This will open a new window. Content with HTML or XML might not display properly. In that case just return to this page.)</i></div>' 
		)
	;
	putlog attributeValue= ;
end ;
else do ;
	select  ;
		when ( attributeName in ( 'Desc' , 'StoredText' ) ) do ;
		put attributeName= attributeValue= attributeOrder= ;
		put attributeValue= ;
		/* Catch any sequence of characters starting with [ and not containing []| until | as $1,
		and the next similar sequence as $2 
		This will turn a wiki-like link [URL|TEXT] into <a href="URL">TEXT</a>  */
			attributeValue = prxchange ( 's/\[([^\[]*?)\|(.*?)\]/<a href="$1">$2<\/a>/', -1 , attributeValue ); 
		/* Change encoded newline to <BR> */
			attributeValue = tranWrd ( attributeValue , '&#xA;' , '<BR>' ) ;
		put attributeValue= ;
		end ;

		otherwise  ;
	end ;
end ;
run ;

/* define the target locations for the pieces of HTML */
filename attribs temp encoding = "utf-8" ;
filename objhdr temp encoding = "utf-8" ;
filename assocs temp encoding = "utf-8" ;
filename trail temp encoding='utf-8';


data _null_ ;
length 
	type  name publicType assocName $ 40 
	trailAssoc $ 22 
	trailId objid $ 17 
	assocId $ 4 
	abstractTrail 
	newtrail 
	newlink 
	encoded $ 4000
	icon $ 40 
;
/* to access the association table with index */
length type association_associationType association_associatedType $ 40  ;

call missing ( type , objId, publicType, name ) ;
file trail ;

itrail = 1 ;
trailAssoc = scan ( "&pw_mde_trail" , 1 , ',' ) ;
newtrail = '' ;
abstractPath = '' ;

putlog trailAssoc=  ;

do while ( trailassoc > '' ) ;

	trailId = scan ( trailAssoc , 1 , '-' ) ;
	assocId = scan ( trailAssoc , 2 , '-' ) ;
	assocName = put ( assocId , $hex2ass. ) ;


	/* get next objectId, if no more it is &pw_mde_object */
	/* not needed */
/*	nextId = scan ( scan ( "&pw_mde_trail" , itrail+1 , ',' ) , 1 , '-' ) ;*/
/*	if nextId = '' then nextId = "&pw_mde_object" ;*/

	putlog / trailAssoc= trailId= assocId= assocName=  ;


	rc = metadata_resolve ( trailId , type , objid ) ;
	putlog trailId=   type= ;

	publicType = '' ;
	rc = metadata_getattr ( trailId , 'publicType' , publicType ) ;
	putlog rc= publicType= ;

	rc = metadata_getattr ( trailId , 'name' , name ) ;
	putlog name= ;

/*	rc = metadata_resolve ( nextId , association_associatedType  , objid ) ;*/
/*	putlog nextId= association_associatedType= ;*/

	association_associationType = assocName ;

	putlog type= association_associationType= association_associatedType= ; 
	set pw_mde.pw_mde_associations key = partner / unique ;
	putlog _IORC_= association_partner= ;
	if _IORC_ ne 0 then association_partner = 'could not be determined' ;

	abstractTrail = cats ( abstractTrail , '&thinsp;' , type , '&thinsp;&lt;' , association_Partner , '/' , assocName , '&gt;' ) ;
	putlog abstractTrail= ;

	/*	insert optional word breaks at uppercase */
	camelPos = anyUpper ( assocName , 2 ) ;
	do while ( camelPos> 0 ) ;
		char = ksubstr ( assocName , camelPos , 1 ) ;
		assocName = kupdate ( assocname , camelPos , 1 , cats ( '<wbr>' , char ) ) ;
		camelPos = camelPos + 6 ;
		camelPos = anyUpper ( assocName , camelPos ) ;
	end ;

	camelPos = anyUpper ( association_partner , 2 ) ;
	do while ( camelPos> 0 ) ;
		char = ksubstr ( association_partner , camelPos , 1 ) ;
		association_partner = kupdate ( association_partner , camelPos , 1 , cats ( '<wbr>' , char ) ) ;
		camelPos = camelPos + 6 ;
		camelPos = anyUpper ( association_partner , camelPos ) ;
	end ;
	putlog 'WBR' / assocname= association_partner= ;

	if publicType =: 'Embedded:' then publicType = ksubstr ( publicType , 10 ) ;

	icon = '' ;
	encoded = '' ;
	len = 0 ;
	set pw_mde.pw_mde_typeDefinitions key = publicType / unique ;
	putlog 'publicType' _IORC_= icon= ;

	/* if not found on PublicType, try ObjectType */
	objectType = type ;
	if _IORC_ ne 0 then set pw_mde.pw_mde_objectTypeIcons key = objectType / unique ;
	putlog 'objectType' _IORC_= icon= ;

	if _IORC_ = 0 then set pw_mde.PW_MDE_icons key = icon / unique ;
	putlog 'icon' _IORC_= len= ;

	/* Create the link to show this screen starting with this object */
	newlink = 
			cats ( "//&_srvname:&_srvport/&_url?_program=&_program" , 
			'&' , "&debugParm" , '&pw_mde_type=' , type , 
			'&pw_mde_object=' , trailId , 
			'&pw_mde_trail=' , newtrail )
	;
	putlog newlink= ;
	put '<a class="btn btn-default object" role="button" href="' newlink '">' ; 
	if _IORC_= 0 then put '<img src="data:image/gif;base64,' encoded '" height="32" width="32" alt="' icon '"/>' ;
	put type ;
	if publicType > '' and publicType ne type then put ' (' publicType +(-1) ')' ;
	put '<br>' name ;
	put '<br><span class="small">' trailID '</span>' ;
	put '</a>' ;

	put 
		'<span class="btn btn-default assoc" >'
		'<b>&lArr;</b>&nbsp;'
		association_partner '<br>'
		assocName 
		+(-1) '&nbsp;<b>&rArr;</b>' 
		'</span>' 
	;

	newtrail = cats ( newtrail , ',' , trailassoc ) ;

	iTrail + 1 ;
	trailassoc = scan ( "&pw_mde_trail" , iTrail , ',' ) ;
	putlog iTrail= trailassoc= ;


end ;
abstractTrail = cats ( abstractTrail , '&nbsp;' , "&pw_mde_type" ) ;
put '<br>&nbsp;<b>Abstracted:</b><tt>&nbsp;' abstractTrail '</tt>'   ;
putlog 'ready' ;
/* STOP needed because a table is accessed with KEY=, otherwise an endless loop would originate */
stop ;

run ;



ods _ALL_ close ;

ods escapechar='~' ;

ods tagsets.bootstrap file = attribs ;

proc report data = attributes ( where = ( attributeOrder >= 10 ) ) nowd ;
columns attributeName attributeValue ;
define attributeValue /
	display "" right 
;
define attributeName /
	display "" left
;
run ;

ods tagsets.bootstrap close ;

ods tagsets.bootstrap file = assocs ;

proc report data = associations nowd ;
by associationClass ;
label associationClass = 'Association' ;
columns associationClass associationId associationType associationName ;
define associationClass / noprint ;
define associationId /
	"Id" display
;
define associationType /
	"Object Type" display 
;
define associationName /
	"Object Name" display
;

endComp ;
compute associationType ; 
	call define ( 2 , "URL" , 
		cats ( "//&_srvname:&_srvport/&_url?_program=&_program" , 
			'&' , "&debugParm" , '&pw_mde_type=' , associationType , 
			'&pw_mde_object=' , associationId , 
			'&pw_mde_trail=' , "&pw_mde_trail" , ",&pw_mde_object" , '-' , put ( associationClass , $ass2hex4. ) 
		) 
	) ;
endComp ;

run ;

ods tagsets.bootstrap close ;

data _null_ ;
length icon $ 40 ;
file objhdr ;
set objInfo ;

	putlog // 'Creating button for selected object' ;

	put '<button class="btn btn-default object " style="font-size: 125%; border-width: 3px; pointer-events: none"  >' ; 

	publicType = objectPublicType ;
	putlog publicType= ;

	icon = '' ;
	if publicType =: 'Embedded:' then publicType = ksubstr ( publicType , 10 ) ;
	set pw_mde.pw_mde_typeDefinitions key = publicType / unique ;
	putlog _IORC_= icon= ;

	/* if not found on PublicType, try ObjectType */
	if _IORC_ ne 0 then set pw_mde.pw_mde_objectTypeIcons key = objectType / unique ;
	putlog _IORC_= icon= ;


	if _IORC_ = 0 then set pw_mde.PW_MDE_icons key = icon / unique ;
	putlog _IORC_= len= ;

	if _IORC_ = 0 then put '<img src="data:image/gif;base64,' encoded '" height="32" width="32" alt="' icon '"/>' ;

	put '<span style="color:gray;">Objecttype: </span>' objectType ;

if objectPublicType > '' and objectPublicType ne objectType then put ' <span style="color:gray;">(PublicType:</span> ' objectPublicType +(-1) ')' ;
else put '<span style="color:gray; font-style:italic; font-size: 80%;">(PublicType not set)</span>' ;
put '<br><span style="color:gray;">Name:</span> ' objectName '<br>' ;
if objectDesc > "" then put '<span style="color:gray;">Desc:</span> ' objectDesc ;
else put '<span style="color:gray; font-style:italic; font-size: 80%;">(no Description)</span>' ;
put '<br><span class="small"><tt>' objectId '</tt></span>' ;
	put '</button>' ;


%macro PW_MDE_CheckObjectCode ;
/* Check for Column and other specials */

%let objectTypeSASfile = %sysfunc ( fileExist ( &pw_mde_libraryPath./PW_MDE_Object_&PW_MDE_type..html ) ) ;
%put &=objectTypeSASfile ;
%if ( &objectTypeSASfile ) %then %do ;
%include addhtml ;
%end ;
%mend /* PW_MDE_CheckObjectCode */ ;

filename addHTML "&pw_mde_libraryPath./PW_MDE_Object_&PW_MDE_type..html"  ;


proc stream outfile = content ; BEGIN
%if ( "&pw_mde_trail" > "" ) %then %do ;
<div class="row" style="padding: 6px;">
&streamdelim readfile trail ;
</div>
%end ;
&streamdelim readfile objhdr ;

%PW_MDE_CheckObjectCode 



<div class="row" style="padding: 6px;">
<div class="col-lg-6">
<h4>Other Attributes</h4>
&streamdelim readfile attribs ;
</div>
<div class="col-lg-6">
<h4>Associations</h4>
&streamdelim readfile assocs ;
</div>
</div>

;;;;

/* append the current object to the trail for the next loop, keep old value for use in display  */

%PW_MDE_ResultPage ;

%mend /* PW_MDE_Info */ ;

%macro PW_MDE_CheckObjectType ;
/* Check for Column and other specials, otherwise */

%let objectTypeSASfile = %sysfunc ( fileExist ( &pw_mde_libraryPath./PW_MDE_Type_&PW_MDE_type..sas ) ) ;
%put &=objectTypeSASfile ;
%if ( &objectTypeSASfile ) %then %include "&pw_mde_libraryPath./PW_MDE_Type_&PW_MDE_type..sas" ;
%else %PW_MDE_Objects ;

%mend /* PW_MDE_CheckObjectType */;


%macro PW_MDE_Attribute ;

%let flag = 67111316;


proc stream outfile=request; BEGIN<GetMetadataObjects>
<Reposid>$METAREPOSITORY</Reposid>
<Type>&pw_mde_type.</Type>
<Objects/>
<NS>SAS</NS>
<Flags>&Flag.</Flags>
<Options>
<Templates>
	<&pw_mde_type. &pw_mde_attribute.="" TextRole="" TextType="" />
</Templates>
<XMLSELECT search="@Id='&pw_mde_object.'"/>
</Options>
</GetMetadataObjects>
;;;;

proc metadata in=request out=response ;
run;

/* Map to get TextType */
proc stream outfile=xmlmap ;BEGIN<?xml version="1.0" encoding="windows-1252"?>
<SXLEMAP name="SXLEMAP" version="2.1">
    <NAMESPACES count="0"/>
    <TABLE name="TextType">
        <TABLE-PATH syntax="XPath">/GetMetadataObjects/Objects/&pw_mde_type.</TABLE-PATH>

        <COLUMN name="TextType">
            <PATH syntax="XPath">/GetMetadataObjects/Objects/&pw_mde_type./@TextType</PATH>
            <TYPE>character</TYPE>
            <DATATYPE>string</DATATYPE>
            <LENGTH>10</LENGTH>
        </COLUMN>
    </TABLE>
</SXLEMAP>
;;;;
libname response xmlv2  xmlmap=xmlmap ;
%let textType =XML;
/* Table might not exist */
%if ( %sysfunc ( exist ( response.textType ) ) ) %then %do ;

data _null_ ;
set response.textType ;
put _ALL_ ;
if textType > "" then call symput ( 'textType' , textType ) ;
run ;

%end ;
%put &=textType ;

/* stylesheet to extract the single attribute */
proc stream outfile=xslt; BEGIN<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="text" />

 <xsl:template match="/">
  <xsl:apply-templates select="GetMetadataObjects/Objects/&pw_mde_type." />
 </xsl:template>

 <xsl:template match="&pw_mde_type.">
  <xsl:value-of select="@&pw_mde_attribute." />
 </xsl:template>

</xsl:stylesheet>
;;;;


filename webout  temp encoding = 'utf-8' ;

/* loop because PROC XSL might not give results, just execute again.
See  http://support.sas.com/kb/62/918.html */
%let rc=not executed ;
%if ( %sysfunc ( fexist ( webout ) ) ) %then %let rc = %sysfunc ( fdelete ( webout ) ) ;
%put &=rc ;
%let loop = 0 ;
%do %while (%sysfunc(fexist(webout)) = 0 and &loop. < 5 );
%let loop = %eval(&loop. + 1);
%put &=loop;
/* stream the attribute value to a file */
proc xsl in=response xsl=xslt out=webout;
run;
%end;

%if ( %sysfunc ( fexist ( webout ) ) ) %then %do ;
/* if the extraction succeded, move it to _webout */
data _null_;
 rc = stpsrv_header('Content-type',"text/&textType" );
run;

proc stream outfile=_webout ; BEGIN
&streamdelim readfile webout ;
;;;;

%end ;

%else %do ;
/* display error message */
proc stream outfile=content ; BEGIN<br>
<p>
We are sorry, but the application was not able to extract the attribute value <code>&pw_mde_attribute</code> for <i>&pw_mde_type</i> object <code>&pw_mde_object</code> succesfully from the metadata.</p>
<p>The value might be too complicated.
</p>
;;;;

%PW_MDE_ResultPage ;

%end ;

%mend ; /*PW_MDE_Atribute*/ 


%macro PW_MDE_PreInit ;

%if &PW_MDE_versionOK %then %do ;
	%put NOTE: Everything OK ;
%end ;
%else %do ;
	%put NOTE: Update necessary, will try that now ;
	%if not %symexist ( _metafolder ) %then %do ;
		%put ERROR%str(:) For this the macrovariable _metafolder has to be defined,;
		%put ERROR%str(:) and should contain the path to the metadata folder with the Stored Process;
	%end ;
	%else 	%PW_MDE_ExecUpdate;
%end ;

%mend ; /* PW_MDE_PreInit / 

/* All macros have been defined. Now select the one to execute */
%macro PW_MDE_SelectAction ;
%if not %symexist ( _CLIENT ) %then %PW_MDE_PreInit ;
%else %if not ( "%scan ( &_CLIENT , 1 )" = "StoredProcessService" ) %then %PW_MDE_PreInit ;
%else %if "&PW_MDE_update" = "yes" %then %PW_MDE_Form ;
%else %if "&PW_MDE_versionOK" ne "1" %then %PW_MDE_PrepUpdate ;
%else %if "&PW_MDE_code" > "" %then %include "&pw_mde_libraryPath./PW_MDE_&PW_MDE_code..sas" ;
%else %if "&PW_MDE_type" = "" %then %PW_MDE_Form ;
%else %if "&PW_MDE_object" > "" and "&PW_MDE_attribute" > "" %then %PW_MDE_Attribute ;
%else %if "&PW_MDE_type" > "" and "&PW_MDE_object" > "" %then %PW_MDE_Info ;
%else %PW_MDE_CheckObjectType ;
%mend PW_MDE_SelectAction ;


options mprint symbolgen ;
%PW_MDE_selectAction ;


